/* $Id: sockl_proc.c,v 1.3 2000/10/10 22:10:44 dps Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <errno.h>
#include "libdetail/socklist.h"
#include "libdetail/detail.h"
#include "cfg_log.h"


static socklist *read_unix(FILE *in, enum __socket_type t);
static socklist *read_ipv4(FILE *in, enum __socket_type t);

static const struct {
    sa_family_t fam;
    enum __socket_type t;
    const char *file;
    socklist (*readproc)(FILE *, enum __socket_type);
} sock_files[]=
{
    AF_INET, SOCK_STREAM, "/proc/net/tcp", &read_ipv4, 
    AF_INET, SOCK_DGRAM,  "/proc/net/udp", &read_ipv4,
    AF_INET, SOCK_RAW, "/proc/net/raw", &read_ipv4,
    AF_UNIX, SOCK_STREAM, "/proc/net/unix", &read_unix
};

#ifndef PATHMAX
#define PATHMAX 2048
#endif /* PATHMAX */


/******************** Common ********************/

/* Socket file name information */
static const char *sockfile(const socklist *p)
{
    int i;

    for (i=sizeof(sock_files)/sizeof(sock_files[0])-1; i>=0; i--)
    {
	if (sock_files[i].fam==p->fam && sock_files[i].t==p->type)
	    return sock_files[i].file;
    }
    return NULL;
}

/* Get socket list */
static socklist *get_socket_list(const char *fname, const enum __socket_type t,
				 socklist *(*read)(FILE *, enum __socket_type))
{
    FILE *in;
    socklist *res, *rd;
    int line, c;

    if ((in=open_procfile(fname))==NULL)
	return NULL;		      /* Open the line */
    if (!eat_line(in))
    {
	do_log("/proc cracking detected: header line far too long in %s", fname);
	fclose(in);
	return NULL;
    }

    res=NULL;
    line=2;
    while ((c=fgetc(in))!=EOF)		    /* Apparetnly more reliable than feof */
    {
	ungetc(c, in);			    /* Push c back */
	if ((rd=(*read)(in, t))==NULL)
	{
	    do_log("/proc cracking detcected: format violated in %s line %d",
		fname, line);
	    eat_line(in);		    /* Eat rest of line */
	}
	else
	{
	    rd->next=res;
	    res=rd;
	}
	line++;

    }
    fclose(in);
    return res;
}



/******************** unix ********************/

/* Read a line of unix socker */
static socklist *read_unix(FILE *in, enum __socket_type t)
{
    int c, i, inode, type, digits;
    ino_t ino;
    struct sockaddr_un *src;
    char s[PATHMAX];
    socklist *res;

    t=t;				    /* Stop compiler warning */

    if (!check_fmt(in, "%8x: %8x %8x %8x "))
	return NULL;

    type=readhex(in, &digits);		    /* Type */

    if (!check_fmt(in," %2x "))
	return NULL;

    inode=readdec(in, &digits);
    if (absorb_ws(in)<=0)
	return NULL;
    for (i=0; (i<sizeof(s)-1 && (c=fgetc(in))!=EOF); i++)
    {
	if (c=='\n')
	    break;
	s[i]=c;
    }
    s[i]='\0';

    if ((res=(socklist *) malloc(sizeof(socklist)))==NULL)
    {
	ungetc('\n',in);		    /* Push back \n */
	return NULL;
    }

    src=(struct sockaddr_un *) malloc(sizeof(struct sockaddr_un));
    if (src==NULL)
    {
	free(res);
	do_log("read_unix: proto %d, inode %d, path %s",
	    type, (int) inode, s);
	ungetc('\n',in);		    /* Push back \n */
	return NULL;
    }
    
    src->sun_family=AF_UNIX;
    strncpy(src->sun_path, s, sizeof(src->sun_path));
    *(src->sun_path+sizeof(src->sun_path)-1)='\0';
    res->fam=AF_UNIX;					/* Family */
    res->type=type;					/* Type */
    res->src=(struct sockaddr *) src;			/* Source address */
    res->dest=NULL;					/* Destination n/a */
    res->dev=UNKNOWN_DEVICE;				/* Unknown device */
    res->inode=inode;					/* Inode */
    res->uid=UNKNOWN_UID;				/* Uknown uid */
    res->refs=NULL;					/* Refs not known */
    return res;
}



/******************** IPv4 ********************/

/* Read a IPv4 address */
static struct sockaddr_in *__ipv4_addr(FILE *in)
{
    unsigned int  addr;		/* Address */
    unsigned int  port;		/* Port number */
    int digits;			/* Digit count */
    struct sockaddr_in *res;	/* Result */
    int c;

    addr=readhex(in, &digits);
    if (digits!=8)
    {
	do_log("Hacked /proc detected: host not 8 digits");
	return NULL;
    }
    if ((c=fgetc(in))!=':')
    {
	ungetc(c, in);
	do_log("Hacked /proc detected: address not followed by : but %d", c);
	return NULL;
    }
    port=readhex(in, &digits);
    if (digits!=4)
    {
	do_log("Hacked /proc detected: port number not 4 digits");
	return NULL;
    }
    if ((res=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
    {
	do_log("System in distress: could not allocate %d bytes",
	    sizeof(struct sockaddr_in));
	return NULL;
    }
/*    res->sa_family=AF_INET;	  */    /* IPv4 */
    res->sin_addr.s_addr=addr;	      /* Address (32 bits) */
    res->sin_port=(uint16_t) port;    /* Port number (16 bits) */
    return res;
}

/* Read a line form /proc/net/tcp, /proc/net/udp or /proc/net/raw for linux 2.2.x */
static socklist *read_ipv4(FILE *in, enum __socket_type t)
{
    socklist *res;
    struct sockaddr_in *src, *dst;
    int uid, inode;
    int c, digits;

    if (!check_fmt(in, "!%d: "))
    {
	do_log("/proc cracking detected: wrong format before local address");
	return NULL;
    }
    /* <local adress><white space><remote address> */
    if ((src=__ipv4_addr(in))==NULL)  /* Source address is intersting */
    {
	do_log("/proc cracking detectded: no source address");
	return NULL;
    }
    if (!absorb_ws(in)>0)
    {
	do_log("/proc cracking detected: junk on end of spurce address");
	free(src);
	return NULL;
    }
    if ((dst=__ipv4_addr(in))==NULL)  /* So is destination address */
    {
	free(src);
	do_log("/proc cracking detected: no destination address");
	return NULL;
    }
    /* Lots of stuff we do not care about */
    if (!check_fmt(in," %2x %8x:%8x %2x:%8x %8x "))
    {
	free(src);
	free(dst);
	do_log("/proc cracking detected: wrong format after destination address");
	return NULL;
    }
    /* <uid><white space><timeout><white space><inode> */
    if ((uid=readdec(in, &digits))==-1)    /* User id is interesting information */
    {
	do_log("/proc cracking detected: uid not present");
	free(src);
	free(dst);
	return NULL;
    }
    if (!check_fmt(in," %d "))
    {
	free(src);
	free(dst);
	do_log("/proc cracking detected: bad format around timeout in %%d format");
	return NULL;
    }
    if ((inode=readdec(in, &digits))==-1)   /* Inode is interesting information */
    {
	do_log("/proc cracking detected: inode not present");
	free(src);
	free(dst);
	return NULL;
    }
    if (!check_eol(in))		      /* Now check end of line */
    {
	do_log("/proc cracking detected: trailing junk");
	free(src);
	free(dst);
	ungetc('\n',in);		    /* Push back \n */
	return NULL;
    }

    /* Whew! Now we have gathered source, destination, uid and inode (and enforced
       the format of the rest) */
    if ((res=(socklist *) malloc(sizeof(socklist)))==NULL)
    {
	do_log("System in distress: could not allocate %d bytes",
	    sizeof(struct sockaddr_in));
	free(src);
	free(dst);
	ungetc('\n',in);		    /* Push back \n */
	return NULL;
    }
    res->fam=AF_INET;			    /* IPv4 */
    res->type=t;			    /* Passed */
    res->src=(struct sockaddr *) src;	    /* Source address */
    res->dest=(struct sockaddr *) dst;	    /* Destination */
    res->dev=UNKNOWN_DEVICE;		    /* dev not avialable here */
    res->inode=inode;			    /* Inode */
    res->uid=uid;			    /* User id */
    res->refs=NULL;			    /* Refs not available here */
    return res;				    /* All done */
}


/* Exported entry point */
socklist *sockl_proc(void)
{
    int i;
    socklist *res, *out;

    res=NULL;
    for (i=0; (i<sizeof(sock_files)/(sizeof(sock_files[0]))); i++)
    {
	out=get_socket_list(sock_files[i].file, sock_files[i].t, sockfiles[i].readproc);
	socklist_append(&res, out);
    }
    return res;
}

/******************** Testing ********************/	

#ifdef TEST
int main(void)
{
    socklist *tcp, *udp, *raw, *unx;

    icache_init();

    tcp=get_socket_list("/proc/net/tcp", SOCK_STREAM, read_ipv4);
    udp=get_socket_list("/proc/net/udp", SOCK_DGRAM, read_ipv4);
    raw=get_socket_list("/proc/net/raw", SOCK_RAW, read_ipv4);
    unx=get_socket_list("/proc/net/unix", SOCK_STREAM, read_unix);
    print_socklist(tcp);
    print_socklist(udp);
    print_socklist(raw);
    print_socklist(unx);
    socklist_zap(tcp);
    socklist_zap(udp);
    socklist_zap(raw);
    socklist_zap(unx);
    return 0;
}
#endif /* TEST */
